/*
 * Turbo Programmer Utilities, turbo-prog-utils, www.bladox.com 
 *
 * Copyright (C) 2004, 2005 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <tprog/tprog.h>
#include <stdlib.h>
#include <avr/interrupt.h>

#include "brum.h"
#include "brumuart.h"

//    actualUbrr = 0;// 1 MBd @ 8 MHz @double speed
//    actualUbrr = 8;// 115k2 @ 8 MHz @double speed
//    actualUbrr = 16;// 57k6 @ 8 MHz @double speed
//    actualUbrr = 51;// 19200 @ 8 MHz @double speed
//    actualUbrr = 103;// 9600 @ 8 MHz @double speed
//    actualUbrr = 207;// 4800 @ 8 MHz @double speed
//    actualUbrr = 0;// 0,5 MBd @ 8 MHz @normal speed
//    actualUbrr = 3;// 115k2 @ 8 MHz @normal speed
//    actualUbrr = 8;// 57k6 @ 8 MHz @normal speed
//    actualUbrr = 25;// 19200 @ 8 MHz @normal speed
//    actualUbrr = 51;// 9600 @ 8 MHz @normal speed
//    actualUbrr = 103;// 4800 @ 8 MHz @normal speed
//#define BRUM_UART_DEFAULT_UBRRL (51)
//#define BRUM_UART_DEFAULT_UBRRL (8)
#define BRUM_UART_DEFAULT_UBRRL (0)
#define BRUM_UART_DEFAULT_UBRRH (0)
u16 actualUbrr;

#define BRUM_TX_SM_STATE1 (1)
#define BRUM_TX_SM_STATE2 (2)
u8 brumUartTxStateMachine;

#define BRUM_UART_RX_FIFO_SIZE 8
u8 brumUartRxFifo[BRUM_UART_RX_FIFO_SIZE];
u8 brumUartRxFifoNumOfChars;
u8 brumUartRxFifoTail;
u8 brumUartRxFifoHead;

void init_brumUart (void)
{
  brumUartRxFifoHead = 0;
  brumUartRxFifoTail = 0;
  brumUartRxFifoNumOfChars = 0;

  actualUbrr = ((BRUM_UART_DEFAULT_UBRRH) << 8) | (BRUM_UART_DEFAULT_UBRRL);

  brumUartTxStateMachine = BRUM_TX_SM_STATE1;

  // clear RXC flag (write log.1 to it)
  sbi (BRUM_UART_UCSRA, RXC);
  // clear TXC flag (write log.1 to it)
  sbi (BRUM_UART_UCSRA, TXC);

  //set mode to double speed
  outb (BRUM_UART_UCSRA, 0x02);
//    //set mode to normal speed
//    outb(BRUM_UART_UCSRA,0x00);

  // enable receive complete interrupt
  // disable transmit complete interrupt
  // enable receiver
  // disable transmitter because of RXD and TXD connected together
  // to not output log.1 to TXD and RXD
  // we enable transmitter before each char will be send
  outb (BRUM_UART_UCSRB, 0x90);

  // asynchronous mode
//  // 8 bit data, even parity, 1 stop bit
//  outb (BRUM_UART_UCSRC, 0x26);
  // 8 bit data, no parity, 1 stop bit
  outb (BRUM_UART_UCSRC, 0x06);

  // set baud rate register to DEFAULT_DIVIDER
  outb (BRUM_UART_UBRRH, (actualUbrr >> 8));
  outb (BRUM_UART_UBRRL, (actualUbrr & 0xFF));
}

void brum_uart_irq ()
{
  // next BRUM UART char transmit will execute delay before transmitting char
  brumUartTxStateMachine = BRUM_TX_SM_STATE1;

  if (brumUartRxFifoNumOfChars == BRUM_UART_RX_FIFO_SIZE)
  {
  }
  else
  {
    brumUartRxFifo[brumUartRxFifoHead] = inb (BRUM_UDR);

    brumUartRxFifoNumOfChars++;
    if (brumUartRxFifoHead < (BRUM_UART_RX_FIFO_SIZE - 1))
      brumUartRxFifoHead++;
    else
      brumUartRxFifoHead = 0;
  }
}

u8 brumUartRxChars (void)
{
  u8 tmpRxChars;

  cli ();
  tmpRxChars = brumUartRxFifoNumOfChars;
  sei ();
  return tmpRxChars;
}

u8 brumUartRxGetChar (void)
{
  u8 tmpTail;

  cli ();
  if (brumUartRxFifoNumOfChars > 0)
  {
    brumUartRxFifoNumOfChars--;
    sei ();
    tmpTail = brumUartRxFifoTail;
    if (brumUartRxFifoTail < (BRUM_UART_RX_FIFO_SIZE - 1))
      brumUartRxFifoTail++;
    else
      brumUartRxFifoTail = 0;

    return brumUartRxFifo[tmpTail];
  }
  else
  {
    sei ();
    return 0;
  }
}

void brumUartTxSendChar (u8 txValue)
{
  u8 txValueTmp;
  u8 iTmp;
  u8 interrupts;

  //store global interrupt enable bit
  interrupts = (inb (SREG) & (1 << SREG_I));

  //cli();
  txValueTmp = txValue;

  if (brumUartTxStateMachine == BRUM_TX_SM_STATE1)
  {
    // execute delay before transmitting char
    //do delay interval
    delayMs (10);
    brumUartTxStateMachine = BRUM_TX_SM_STATE2;
  }

//  delayMs (1);

  // disable receive complete interrupt
  // disable receiver
  outb (BRUM_UART_UCSRB, (inb (BRUM_UART_UCSRB) & 0x6F));

  //TXD1 pin to log.1
  sbi (BRUM_TXD_PORT, BRUM_TXD_PORT_PIN);

  //TXD1 pin change direction to output only when transmitting char
  sbi (BRUM_TXD_DDR, BRUM_TXD_DDR_DD);

  // enable transmitter
  outb (BRUM_UART_UCSRB, (inb (BRUM_UART_UCSRB) | 0x08));

  // send txValueTmp to UDR register
  sbi (BRUM_UART_UCSRA, TXC);
  outb (BRUM_UDR, txValueTmp);
  loop_until_bit_is_set (BRUM_UART_UCSRA, TXC);
  sbi (BRUM_UART_UCSRA, TXC);

  // disable transmitter
  outb (BRUM_UART_UCSRB, (inb (BRUM_UART_UCSRB) & 0xF7));

  // set TXD1 pin to log.1
  sbi (BRUM_TXD_PORT, BRUM_TXD_PORT_PIN);

  //TXD1 pin change direction to input to allow receiving
  cbi (BRUM_TXD_DDR, BRUM_TXD_DDR_DD);

  // clear TXC flag (write log.1 to it)
  sbi (BRUM_UART_UCSRA, TXC);

  //restore global interrupt enable bit
  if (interrupts != 0)
    sei ();

  // enable receive complete interrupt
  // enable receiver
  outb (BRUM_UART_UCSRB, (inb (BRUM_UART_UCSRB) | 0x90));
//  outb (BRUM_UART_UCSRB, 0x90);
}

//-------------------------------------------------
// Transmit data to BRUM UART
//-------------------------------------------------
void brumUartTxSendData (u8 * data, u16 len)
{
  u16 i;

  for (i = 0; i < len; i++)
    brumUartTxSendChar (data[i]);
}
